import shylock.manager
from shylock.backends import ShylockAsyncBackend
from shylock.exceptions import ShylockException


class Lock:
    """
    The best way to use Shylock. Stores references to backend and name for convenient use.

    >>> lock = Lock("my-lock")
    >>> try:
    >>>     await lock.acquire()
    >>> finally:
    >>>     await lock.release()
    >>> async with lock:
    >>>     print("Locked")
    >>> print("Released")
    """

    def __init__(self, name: str, backend: ShylockAsyncBackend = None):
        self.name = name
        self._backend = shylock.manager.BACKEND if backend is None else backend

        if not issubclass(self._backend.__class__, ShylockAsyncBackend):
            raise ShylockException(
                "shylock.aio.Lock requires a ShylockAsyncBackend, did you mean to use shylock.Lock?"
            )

        self._locked = False
        if self._backend is None:
            raise ShylockException(
                f"No Shylock backend set, configure one with shylock.set_default, "
                f"or pass instace of shylock.ShylockBackend as argument to Lock()"
            )

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.release()

    async def acquire(self, block: bool = True) -> bool:
        """
        Try to acquire the lock - optionally block until available
        :param block: Wait until lock is available
        :return: If lock was successfully acquired - always True if blocking
        """
        res = await self._backend.acquire(self.name, block)

        if res:
            self._locked = True

        return res

    async def locked(self) -> bool:
        """
        Does the lock believe it's currently locked - does not check actual backend
        :return: Locked state
        """
        return self._locked

    async def release(self):
        """
        Release the lock
        """
        if not self._locked:
            raise ShylockException(
                f"Trying to unlock {self.name} without locking it first."
            )
        await self._backend.release(self.name)
        self._locked = False



from shylock.aio.lock import Lock as AsyncLock
from shylock.backends.motorasyncio import ShylockMotorAsyncIOBackend
from shylock.backends.pymongo import ShylockPymongoBackend
from shylock.exceptions import *
from shylock.lock import Lock
from shylock.manager import configure



from time import sleep
from typing import Optional

try:
    from pymongo import MongoClient
    from pymongo.collection import Collection
    from pymongo.database import Database
    from pymongo.errors import DuplicateKeyError, WriteError
except ImportError:
    MongoClient = None
    Collection = None
    Database = None
    DuplicateKeyError = None
    WriteError = None

from shylock.backends import ShylockSyncBackend
from shylock.exceptions import ShylockException

DOCUMENT_TTL = 60 * 5  # 5min seems like a reasonable TTL
POLL_DELAY = 1 / 16  # Some balance between high polling and high delay


class ShylockPymongoBackend(ShylockSyncBackend):
    @staticmethod
    def create(
        client: MongoClient, db: str, collection_name: str = "shylock"
    ) -> "ShylockPymongoBackend":
        """"
        Create and initialize the backend
        :param client: Connected Pymongo client instance
        :param db: The name of the DB to use for locks
        :param collection_name: The name of the collection reserved for shylock
        """
        inst = ShylockPymongoBackend(client, db, collection_name)
        inst._init_collection()
        return inst

    def acquire(self, name: str, block: bool = True) -> bool:
        """
        Try to acquire a lock, potentially wait until it's available
        :param name: Name of the lock
        :param block: Wait for lock
        :return: If lock was successfully acquired - always True if block is True
        """
        doc = {"name": name, "ttl": DOCUMENT_TTL}

        while True:
            try:
                self._coll.insert_one(doc)
                return True
            except DuplicateKeyError:
                if not block:
                    return False
                sleep(POLL_DELAY)
            except WriteError as e:
                # Maybe this should check for blocking? Kinda not related though.
                delay = self._check_retry_exception(e)
                if delay is not None:
                    sleep(delay)
                    continue

                raise

    def release(self, name: str):
        """
        Release a given lock
        :param name: Name of the lock
        """
        while True:
            try:
                self._coll.delete_one({"name": name})
                return
            except WriteError as e:
                delay = self._check_retry_exception(e, 0.25)
                if delay is not None:
                    sleep(delay)
                    continue

                raise

    @staticmethod
    def _check():
        if MongoClient is None:
            raise ShylockException(
                "No pymongo driver available. Cannot use Shylock with Pymongo backend without it."
            )

    def __init__(self, client: MongoClient, db: str, collection_name: str = "shylock"):
        self._check()
        self._client: MongoClient = client
        self._db: Optional[Database] = None
        self._db_name: str = db
        self._coll: Optional[Collection] = None
        self._collection_name: str = collection_name

    def _init_collection(self):
        """
        Ensure the collection is ready for our use
        """
        self._db = self._client[self._db_name]
        self._coll = self._db[self._collection_name]

        self._init_index("name", unique=True)
        self._init_index("_ts", expireAfterSeconds=DOCUMENT_TTL)

    def _init_index(self, index_name: str, **params):
        """
        Set up the given index
        :param index_name:
        :param params: https://api.mongodb.com/python/current/api/pymongo/collection.html
        :return:
        """
        idx_info = self._coll.index_information()
        index_found = False
        for name in idx_info:
            keys = [i[0] for i in idx_info[name]["key"]]

            if index_name in keys:
                index_found = True

        if not index_found:
            self._coll.create_index(index_name, **params)

    @staticmethod
    def _check_retry_exception(
        e: WriteError, default_retry_time=None
    ) -> Optional[float]:
        """
        Check if the given WriteError implies we should just retry a bit later
        :param e:
        :param default_retry_time:
        :return: Time to wait for, or None if shouldn't retry
        """
        delay = None
        if e.code == 16500:  # Azure CosmosDB rate limiting
            delay = default_retry_time  # Delete does nto get a RetryAfterMs
            msg = e.details.get("errmsg", "")
            for part in msg.replace(",", "").split(" "):
                if part.startswith("RetryAfterMs="):
                    delay = float(part[len("RetryAfterMs=") :]) / 1000

        return delay



from typing import Optional, Union

from shylock.backends import ShylockAsyncBackend, ShylockSyncBackend

BACKEND: Optional[Union[ShylockAsyncBackend, ShylockSyncBackend]] = None


def configure(backend: Union[ShylockAsyncBackend, ShylockSyncBackend]):
    """
    Configure default backend to use
    :param backend: The ready to use backend
    """
    global BACKEND
    BACKEND = backend



# -*- coding: utf-8 -*-
from io import open
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="shylock",
    version="1.1.0",
    description="Distributed locks in Python, similar to https://github.com/vaidik/sherlock - also with asyncio support",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/lietu/shylock",
    author="Janne Enberg",
    author_email="janne.enberg@lietu.net",
    packages=["shylock", "shylock.backends", "shylock.aio"],
    keywords="distributed locking lock asyncio",
    python_requires=">=3.6,<4",
    extras_require={"motor": ["motor~=2.0.0"], "pymongo": ["pymongo~=3.10.1"],},
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    project_urls={
        "Bug Reports": "https://github.com/lietu/shylock/issues",
        "Source": "https://github.com/lietu/shylock/",
    },
)



class ShylockAsyncBackend:
    @staticmethod
    def _check():
        raise NotImplementedError()

    async def acquire(self, name: str, block: bool = True):
        raise NotImplementedError()

    async def release(self, name: str):
        raise NotImplementedError()


class ShylockSyncBackend:
    @staticmethod
    def _check():
        raise NotImplementedError()

    def acquire(self, name: str, block: bool = True):
        raise NotImplementedError()

    def release(self, name: str):
        raise NotImplementedError()
